@@ -32,6 +32,19 @@
 
 #if defined(HAVE_GETRLIMIT) && defined(RLIM_INFINITY)
 
+#if defined(HAVE_RLIMIT_POSIXLOCKS) && !defined(HAVE_RLIMIT_LOCKS)
+#  define RLIMIT_LOCKS		RLIMIT_POSIXLOCKS
+#  define HAVE_RLIMIT_LOCKS     1
+#endif
+
+#if defined(HAVE_RLIMIT_NTHR) && !defined(HAVE_RLIMIT_PTHREAD)
+#  define RLIMIT_PTHREAD	RLIMIT_NTHR
+#  define HAVE_RLIMIT_PTHREAD   1
+#  define THREAD_FMT            "-T: threads                         "
+#else
+#  define THREAD_FMT            "-T: threads per process             "
+#endif
+
 enum {
     ZLIMTYPE_MEMORY,
     ZLIMTYPE_NUMBER,
@@ -238,32 +251,32 @@ printulimit(char *nam, int lim, int hard, int head)
     switch (lim) {
     case RLIMIT_CORE:
 	if (head)
-	    printf("-c: core file size (blocks)    ");
+	    printf("-c: core file size (blocks)         ");
 	if (limit != RLIM_INFINITY)
 	    limit /= 512;
 	break;
     case RLIMIT_DATA:
 	if (head)
-	    printf("-d: data seg size (kbytes)     ");
+	    printf("-d: data seg size (kbytes)          ");
 	if (limit != RLIM_INFINITY)
 	    limit /= 1024;
 	break;
     case RLIMIT_FSIZE:
 	if (head)
-	    printf("-f: file size (blocks)         ");
+	    printf("-f: file size (blocks)              ");
 	if (limit != RLIM_INFINITY)
 	    limit /= 512;
 	break;
 # ifdef HAVE_RLIMIT_SIGPENDING
     case RLIMIT_SIGPENDING:
 	if (head)
-	    printf("-i: pending signals            ");
+	    printf("-i: pending signals                 ");
 	break;
 # endif
 # ifdef HAVE_RLIMIT_MEMLOCK
     case RLIMIT_MEMLOCK:
 	if (head)
-	    printf("-l: locked-in-memory size (kb) ");
+	    printf("-l: locked-in-memory size (kbytes)  ");
 	if (limit != RLIM_INFINITY)
 	    limit /= 1024;
 	break;
@@ -273,7 +286,7 @@ printulimit(char *nam, int lim, int hard, int head)
 # if defined(HAVE_RLIMIT_RSS) && !defined(RLIMIT_VMEM_IS_RSS) && !defined(RLIMIT_RSS_IS_AS)
     case RLIMIT_RSS:
 	if (head)
-	    printf("-m: resident set size (kbytes) ");
+	    printf("-m: resident set size (kbytes)      ");
 	if (limit != RLIM_INFINITY)
 	    limit /= 1024;
 	break;
@@ -281,7 +294,7 @@ printulimit(char *nam, int lim, int hard, int head)
 # if defined(HAVE_RLIMIT_VMEM) && defined(HAVE_RLIMIT_RSS) && defined(RLIMIT_VMEM_IS_RSS)
     case RLIMIT_VMEM:
 	if (head)
-	    printf("-m: memory size (kb)           ");
+	    printf("-m: memory size (kbytes)            ");
 	if (limit != RLIM_INFINITY)
 	    limit /= 1024;
 	break;
@@ -289,35 +302,35 @@ printulimit(char *nam, int lim, int hard, int head)
 # ifdef HAVE_RLIMIT_NOFILE
     case RLIMIT_NOFILE:
 	if (head)
-	    printf("-n: file descriptors           ");
+	    printf("-n: file descriptors                ");
 	break;
 # endif /* HAVE_RLIMIT_NOFILE */
 # ifdef HAVE_RLIMIT_MSGQUEUE
     case RLIMIT_MSGQUEUE:
 	if (head)
-	    printf("-q: bytes in POSIX msg queues  ");
+	    printf("-q: bytes in POSIX msg queues       ");
 	break;
 # endif
     case RLIMIT_STACK:
 	if (head)
-	    printf("-s: stack size (kbytes)        ");
+	    printf("-s: stack size (kbytes)             ");
 	if (limit != RLIM_INFINITY)
 	    limit /= 1024;
 	break;
     case RLIMIT_CPU:
 	if (head)
-	    printf("-t: cpu time (seconds)         ");
+	    printf("-t: cpu time (seconds)              ");
 	break;
 # ifdef HAVE_RLIMIT_NPROC
     case RLIMIT_NPROC:
 	if (head)
-	    printf("-u: processes                  ");
+	    printf("-u: processes                       ");
 	break;
 # endif /* HAVE_RLIMIT_NPROC */
 # if defined(HAVE_RLIMIT_VMEM) && (!defined(HAVE_RLIMIT_RSS) || !defined(RLIMIT_VMEM_IS_RSS))
     case RLIMIT_VMEM:
 	if (head)
-	    printf("-v: virtual memory size (kb)   ");
+	    printf("-v: virtual memory size (kbytes)    ");
 	if (limit != RLIM_INFINITY)
 	    limit /= 1024;
 	break;
@@ -325,7 +338,7 @@ printulimit(char *nam, int lim, int hard, int head)
 # if defined HAVE_RLIMIT_AS && !defined(RLIMIT_VMEM_IS_AS)
     case RLIMIT_AS:
 	if (head)
-	    printf("-v: address space (kb)         ");
+	    printf("-v: address space (kbytes)          ");
 	if (limit != RLIM_INFINITY)
 	    limit /= 1024;
 	break;
@@ -333,13 +346,13 @@ printulimit(char *nam, int lim, int hard, int head)
 # ifdef HAVE_RLIMIT_LOCKS
     case RLIMIT_LOCKS:
 	if (head)
-	    printf("-x: file locks                 ");
+	    printf("-x: file locks                      ");
 	break;
 # endif /* HAVE_RLIMIT_LOCKS */
 # ifdef HAVE_RLIMIT_AIO_MEM
     case RLIMIT_AIO_MEM:
 	if (head)
-	    printf("-N %2d: AIO locked-in-memory (kb) ", RLIMIT_AIO_MEM);
+	    printf("-N %2d: AIO locked-in-memory (kbytes)", RLIMIT_AIO_MEM);
 	if (limit != RLIM_INFINITY)
 	    limit /= 1024;
 	break;
@@ -347,44 +360,62 @@ printulimit(char *nam, int lim, int hard, int head)
 # ifdef HAVE_RLIMIT_AIO_OPS
     case RLIMIT_AIO_OPS:
 	if (head)
-	    printf("-N %2d: AIO operations          ", RLIMIT_AIO_OPS);
+	    printf("-N %2d: AIO operations               ", RLIMIT_AIO_OPS);
 	break;
 # endif /* HAVE_RLIMIT_AIO_OPS */
 # ifdef HAVE_RLIMIT_TCACHE
     case RLIMIT_TCACHE:
 	if (head)
-	    printf("-N %2d: cached threads          ", RLIMIT_TCACHE);
+	    printf("-N %2d: cached threads               ", RLIMIT_TCACHE);
 	break;
 # endif /* HAVE_RLIMIT_TCACHE */
 # ifdef HAVE_RLIMIT_SBSIZE
     case RLIMIT_SBSIZE:
 	if (head)
-	    printf("-N %2d: socket buffer size (kb) ", RLIMIT_SBSIZE);
-	if (limit != RLIM_INFINITY)
-	    limit /= 1024;
+	    printf("-b: socket buffer size (bytes)      ");
 	break;
 # endif /* HAVE_RLIMIT_SBSIZE */
 # ifdef HAVE_RLIMIT_PTHREAD
     case RLIMIT_PTHREAD:
 	if (head)
-	    printf("-N %2d: threads per process     ", RLIMIT_PTHREAD);
+	    printf("%s", THREAD_FMT);
 	break;
 # endif /* HAVE_RLIMIT_PTHREAD */
 # ifdef HAVE_RLIMIT_NICE
     case RLIMIT_NICE:
 	if (head)
-	    printf("-e: max nice                   ");
+	    printf("-e: max nice                        ");
 	break;
 # endif /* HAVE_RLIMIT_NICE */
 # ifdef HAVE_RLIMIT_RTPRIO
     case RLIMIT_RTPRIO:
 	if (head)
-	    printf("-r: max rt priority            ");
+	    printf("-r: max rt priority                 ");
 	break;
 # endif /* HAVE_RLIMIT_RTPRIO */
+# ifdef HAVE_RLIMIT_NPTS
+    case RLIMIT_NPTS:
+	if (head)
+	    printf("-p: pseudo-terminals                ");
+	break;
+# endif /* HAVE_RLIMIT_NPTS */
+# ifdef HAVE_RLIMIT_SWAP
+    case RLIMIT_SWAP:
+	if (head)
+	    printf("-w: swap size (kbytes)              ");
+	if (limit != RLIM_INFINITY)
+	    limit /= 1024;
+	break;
+# endif /* HAVE_RLIMIT_SWAP */
+# ifdef HAVE_RLIMIT_KQUEUES
+    case RLIMIT_KQUEUES:
+	if (head)
+	    printf("-k: kqueues                         ");
+	break;
+# endif /* HAVE_RLIMIT_KQUEUES */
     default:
 	if (head)
-	    printf("-N %2d:                         ", lim);
+	    printf("-N %2d:                              ", lim);
 	break;
     }
     /* display the limit */
@@ -487,7 +518,7 @@ do_limit(char *nam, int lim, rlim_t val, int hard, int soft, int set)
 		}
 	    } else
 		limits[lim].rlim_cur = val;
-	    if (set && zsetlimit(lim, "limit"))
+	    if (set && zsetlimit(lim, nam))
 		return 1;
 	}
     }
@@ -776,16 +807,21 @@ bin_ulimit(char *name, char **argv, UNUSED(Options ops), UNUSED(int func))
 		case 'c':
 		    res = RLIMIT_CORE;
 		    break;
-# ifdef HAVE_RLIMIT_RSS
-		case 'm':
-		    res = RLIMIT_RSS;
+# ifdef HAVE_RLIMIT_SBSIZE
+		case 'b':
+		    res = RLIMIT_SBSIZE;
 		    break;
-# endif /* HAVE_RLIMIT_RSS */
+# endif /* HAVE_RLIMIT_SBSIZE */
 # ifdef HAVE_RLIMIT_MEMLOCK
 		case 'l':
 		    res = RLIMIT_MEMLOCK;
 		    break;
 # endif /* HAVE_RLIMIT_MEMLOCK */
+# ifdef HAVE_RLIMIT_RSS
+		case 'm':
+		    res = RLIMIT_RSS;
+		    break;
+# endif /* HAVE_RLIMIT_RSS */
 # ifdef HAVE_RLIMIT_NOFILE
 		case 'n':
 		    res = RLIMIT_NOFILE;
@@ -829,6 +865,33 @@ bin_ulimit(char *name, char **argv, UNUSED(Options ops), UNUSED(int func))
 		case 'r':
 		    res = RLIMIT_RTPRIO;
 		    break;
+# else
+#  ifdef HAVE_RLIMIT_NTHR
+		    /* For compatibility with sh on NetBSD */
+		case 'r':
+		    res = RLIMIT_NTHR;
+		    break;
+#  endif /* HAVE_RLIMIT_NTHR */
+# endif
+# ifdef HAVE_RLIMIT_NPTS
+		case 'p':
+		    res = RLIMIT_NPTS;
+		    break;
+# endif
+# ifdef HAVE_RLIMIT_SWAP
+		case 'w':
+		    res = RLIMIT_SWAP;
+		    break;
+# endif
+# ifdef HAVE_RLIMIT_KQUEUES
+		case 'k':
+		    res = RLIMIT_KQUEUES;
+		    break;
+# endif
+# ifdef HAVE_RLIMIT_PTHREAD
+		case 'T':
+		    res = RLIMIT_PTHREAD;
+		    break;
 # endif
 		default:
 		    /* unrecognised limit */
